'use strict';

const _ = require('lodash');
const GoogleCloudBigQuery = require('@google-cloud/bigquery');
const squel = require("squel");
const util = require('util');
const asyncp = require('paradox-asyncp');

const Base = require('./base');
const Model = require('../model');
const Types = require('paradox-utils').SchemaTypes;
const moment = require('moment');

const SqlString = require('sqlstring');

class BigQuery extends Base {

    /* -------------------------------------------------------------------------
     * Link
     * ---------------------------------------------------------------------- */

    getNamespaceName(options) {
        options = options || {};

        return options.namespace || this.options.namespace;
    }

    async namespace(options) {
        let ns = this.getNamespaceName(options);

        let bq = this.getBigQuery();
        let dataset = bq.dataset(ns);

        let [exists] = await dataset.exists();

        if(!exists) {
            let createOpts = _.defaults({}, options.raw, {
                location: options.location || this.options.location,
            });

            await dataset.create(createOpts);
        }

        return dataset;
    }

    getNamespace(options) {
        options = options || {};

        if(!this.namespaces) {
            this.namespaces = {};
        }

        let ns = this.getNamespaceName(options);

        if(!(ns in this.namespaces)) {
            this.namespaces[ns] = this.namespace(options);
        }

        return this.namespaces[ns];
    }

    async table(options) {
        options = options || {};

        let tableName = options.table;
        let dataset = await this.getNamespace(options);
        return dataset.table(tableName);
    }

    async getTable(options) {
        options = options || {};

        if(!this.tables) {
            this.tables = {};
        }

        let tableName = options.table;

        if(!(tableName in this.tables)) {
            this.tables[tableName] = this
                .getNamespace(options)
                .then((dataset) => {
                    let table = dataset.table(tableName);

                    return table
                        .exists()
                        .then((data) => {
                            let exists = data[0];

                            if(exists) {
                                return table;
                            }

                            if(this.options.sync.force) {
                                return this.sync(options);
                            }

                            return null;
                        });
                });
        }

        return this.tables[tableName];
    }

    getBigQuery() {
        if(!this.bigquery) {
            let opts = _.defaults({}, this.options.raw, {
                location: 'EU',
                autoRetry: true,
                maxRetries: 3,

            }, this.options.credentials);

            this.bigquery = GoogleCloudBigQuery(opts);
        }

        return this.bigquery;
    }

    getLink() {
        if(!this.link) {
            this.link = new Promise((resolve, reject) => {
                try {
                    resolve(this.getBigQuery());

                } catch(err) {
                    reject(err);
                }
            });
        }

        return this.link;
    }

    /* -------------------------------------------------------------------------
     * Sync
     * ---------------------------------------------------------------------- */

    async destroy(options) {
        let table = await this.getTable(options);

        if(!table) {
            return null;
        }

        let [result] = await table.delete();

        return result;
    }

    async exists(options) {
        let table = await this.getTable(options);

        if(!table) {
            return false;
        }

        let [exists] = await table.exists();
        return exists;
    }

    createSchemaFields(options) {
        let fields = this.model.getFields();
        let schema = [];

        for(let key in fields) {
            let field = fields[key];
            let type = 'STRING';
            let mode = 'NULLABLE';

            let typeName = field.type.getName();

            switch(typeName) {
                case 'STRING':
                    type = 'STRING';
                    break;
                case 'INTEGER':
                case 'NUMBER':
                    type = 'INTEGER';
                    break;
                case 'FLOAT':
                    type = 'FLOAT';
                    break;
                case 'BOOL':
                case 'BOOLEAN':
                    type = 'BOOLEAN';
                    break;
                case 'TIMESTAMP':
                    type = 'TIMESTAMP';
                    break;
                case 'DATE':
                    type = 'DATE';
                    break;
                case 'DATETIME':
                    type = 'DATETIME';
                    break;
                case 'TIME':
                    type = 'TIME';
                    break;
            }

            if('allowNull' in field) {
                if(field['allowNull']) {
                    mode = 'NULLABLE';
                } else {
                    mode = 'REQUIRED';
                }

            } else {
                mode = 'NULLABLE'; // TODO: hacer que allowNull siempre sea por defecto true,
            }

            schema.push({
                name: key,
                type: type,
                mode: mode,
            });
        }

        return schema;
    }

    async sync(options) {
        let tableName = options.table;
        let schema = this.createSchemaFields(options);

        let dataset = await this.getNamespace(options);
        let table = dataset.table(tableName);

        let [exists] = await table.exists();

        if(exists) {
            return table;
        }

        delete this.tables[tableName];

        let partitionKey = this.model.getPartitionKey();
        let partition;

        if(partitionKey) {
            partition = {
                type: 'DAY',
                field: partitionKey,
            };
        }

        let createOpts = _.defaults({}, options.raw, {
            schema,
            timePartitioning: partition
        });
        await table.create(createOpts);

        return table;
    }

    /* -------------------------------------------------------------------------
     * Insert
     * ---------------------------------------------------------------------- */

    async count(options) {
        options = _.defaults({
            attributes: [
                'count(*) count',
            ],
            useLegacySql: false,

        }, options);

        let query = this.createQueryFromOptions(options);

        let rows = await this.query(query, options);
        let first = _.first(rows);

        if(!first) {
            return 0;
        }

        return first.count;
    }

    fixRow(row) {
        let fields = this.model.getFields();

        for(let key in fields) {
            let field = fields[key];

            let type = field.type.getName();

            if(!(key in row) || !row[key]) {
                continue;
            }

            if(_.isDate(row[key])) {
                if(type == 'DATETIME') {
                    row[key] = moment(row[key]).utc().format('YYYY-MM-DDTHH:mm:ss.SSSSSS');

                } else if(type == 'DATE') {
                    row[key] = moment(row[key]).format('YYYY-MM-DD');
                }
            }

            if(type == 'JSON') {
                row[key] = JSON.stringify(row[key]);
            }
        }

        return row;
    }

    async insert(row, options) {
        let rows = await this.bulkInsert([row], options);
        return rows[0];
    }

    async bulkInsert(rows, options) {
        options = _.defaults({}, options, {
            autoCreate: true,
        });

        let table = await this.table(options);
        rows.forEach(this.fixRow.bind(this));

        let insertOpts = _.defaults({}, options.raw, {
            autoCreate: options.autoCreate,
            schema: this.createSchemaFields(options),
        });

        await table.insert(rows, insertOpts);

        return rows;
    }

    update(data, options) {
        for(let key in options.where) {
            data[key] = options.where[key];
        }

        return this.insert(data, options);
    }

    /* -------------------------------------------------------------------------
     * Find
     * ---------------------------------------------------------------------- */

    async metadata(options) {
        let table = await this.table(options);

        if(!table) {
            return null;
        }

        let [metadata] = await table.getMetadata();

        return metadata;
    }

    async explain(query, options) {
        options = _.defaults({}, options, {
            useLegacySql: true,
        });

        return this.query(query, _.defaults({explain: true}, options));
    }

    async query(query, options) {
        options = _.defaults({}, options, {
            useLegacySql: true,
            explain: false,
            replacements: null,
            raw: {},
            jobRaw: {},
            maxResults: undefined,
            timeout: null,
        });

        if(_.isObject(query)) {
            query = this.createQueryFromOptions(_.defaults({}, query, options));
        }

        if(query.match(/\/\*\s*useLegacySql\s*\*\//i)) {
            options.useLegacySql = true;

        } else if(query.match(/\/\*\s*useStandardSql\s*\*\//i)) {
            options.useLegacySql = false;
        }

        query = query.replace(/%PROJECT%/g, this.options.credentials.projectId);
        query = query.replace(/%NAMESPACE%/g, this.getNamespaceName(options));
        query = query.replace(/%TABLE%/g, options.table);

        if(options.replacements) {
            if(_.isArray(options.replacements)) {
                query = SqlString.format(query, options.replacements);

            } else {
                for(let key in options.replacements) {
                    let regex = new RegExp(':' + key, 'gi');
                    query = query.replace(regex, SqlString.escape(options.replacements[key]));
                }
            }
        }

        let dataset = await this.getNamespace(options);

        let queryOpts = _.defaults({}, options.raw, {
            query,
            useLegacySql: options.useLegacySql,
            dryRun: options.explain,
        });

        let [job] = await dataset.createQueryJob(queryOpts);

        if(queryOpts.dryRun) {
            return job.metadata;
        }

        let jobOpts = _.defaults({}, options.jobRaw, {
            maxResults: options.maxResults || undefined,
            timeoutMs: options.timeout || undefined,
        });

        let [rows] = await job.getQueryResults(jobOpts);

        // let [rows] = await dataset.query(queryOpts);

        return this.parseResult(rows);
    }

    async findAll(options) {
        options = _.defaults({}, options, {
            useLegacySql: true,
        });

        let query = this.createQueryFromOptions(options);

        return await this.query(query, options);

        // let promise = this
        //     .getTable(options)
        //     .then(table => {
        //         return table
        //             .query({
        //                 query,
        //             })
        //             .then(data => {
        //                 let rows = data[0];
        //                 return this.parseResult(rows);
        //             });
        //     });
        //
        // this.initiateIterator(promise, query, options);
        //
        // return promise;
    }

    // initiateIterator(promise, query, options) {
    //     promise.iterateCallback = null;
    //
    //     promise.iterate = (callback) => {
    //         return new Promise((resolveIt, rejectIt) => {
    //             this
    //                 .getTable()
    //                 .then(table => {
    //                     let procressResults = (err, rows, nextQuery) => {
    //                         if(err) return rejectIt(err);
    //
    //                         let chunks = 1000;
    //
    //                         if(_.isNumber(options.iterate)) {
    //                             chunks = options.iterate;
    //                         }
    //
    //                         asyncp
    //                             .eachLimit(_.chunk(rows, chunks), 1, subRows => {
    //                                 if(promise.iterateCallback) {
    //                                     subRows = subRows.map(promise.iterateCallback);
    //                                 }
    //
    //                                 let res = callback.apply(this, [subRows]);
    //
    //                                 if(!this.isPromise(res)) {
    //                                     res = Promise.resolve();
    //                                 }
    //
    //                                 return res
    //                             })
    //                             .then(() => {
    //                                 if(nextQuery) {
    //                                     table.query(nextQuery, procressResults);
    //
    //                                 } else {
    //                                     resolveIt();
    //                                 }
    //                             })
    //                             .catch(rejectIt);
    //                     };
    //
    //                     table.query({
    //                         query: query,
    //                         autoPaginate: false,
    //                     }, procressResults);
    //                 });
    //
    //         });
    //     };
    // }

    parseResult(rows) {
        _.forEach(rows, row => {
            _.forEach(row, (value, key) => {
                if(value && _.isObject(value) && 'value' in value && value['value'].match(/^\d+-/)) {
                    value = new Date(value['value']);
                    row[key] = value;
                }
            });
        });

        return rows;
    }

    /* -------------------------------------------------------------------------
     * Utils
     * ---------------------------------------------------------------------- */

    createQueryFromOptions(options) {
        let query = squel.select();

        let tableName = this.tableName(options.table, options);
        query.from(tableName);

        /**
         * @deprecated use attributes
         */
        if(options.fields) {
            options.attributes = options.fields;
        }

        if(options.attributes) {
            this.applyFields(query, options.attributes);
        }

        if(options.where) {
            this.applyWhere(query, options.where);
        }

        if(options.order) {
            this.applyOrder(query, options.order);
        }

        if(options.limit) {
            this.applyLimit(query, options.limit);
        }

        if(options.offset) {
            this.applyOffset(query, options.offset);
        }

        return query.toString();
    }

    applyFields(query, fields) {
        fields.forEach(field => {
            query.field(field);
        });
        return this;
    }

    applyWhere(query, where) {
        for(let key in where) {
            let value = where[key];

            if(typeof value == 'object') {
                for(let operator in value) {
                    let subvalue = value[operator];

                    if(operator == '$raw') {
                        query.where(key + ' ' + subvalue);

                    } else {
                        let compatOperator = this.convertOperator(operator);

                        if(_.isDate(subvalue)) {
                            subvalue = "CAST('" + moment(subvalue).format('YYYY-MM-DD HH:mm:ss') + "' AS DATETIME)";
                            query.where(util.format(compatOperator.replace(/\?/g, '%s'), key, subvalue));

                        } else {
                            query.where(util.format(compatOperator, key), subvalue);
                        }

                    }
                }

            } else {
                query.where(util.format('%s %s ?', key, '='), value);
            }
        }

        return this;
    }

    applyOrder(query, order) {
        return this;
    }

    applyLimit(query, limit) {
        query.limit(limit);
        return this;
    }

    applyOffset(query, offset) {
        query.offset(offset);
        return this;
    }

    tableName(name, options) {
        let namespace = options.namespace || this.options.namespace;

        if(options.useLegacySql) {
            name = '[' + this.options.credentials.projectId + ':' + namespace + '.' + name + ']';

        } else {
            name = '`' + this.options.credentials.projectId + '.' + namespace + '.' + name + '`';
        }

        return name;
    }

    convertOperator(operator) {
        /*
         $and: {a: 5}           // AND (a = 5)
         $or: [{a: 5}, {a: 6}]  // (a = 5 OR a = 6)
         $gt: 6,                // > 6
         $gte: 6,               // >= 6
         $lt: 10,               // < 10
         $lte: 10,              // <= 10
         $ne: 20,               // != 20
         $eq: 3,                // = 3
         $not: true,            // IS NOT TRUE
         $between: [6, 10],     // BETWEEN 6 AND 10
         $notBetween: [11, 15], // NOT BETWEEN 11 AND 15
         $in: [1, 2],           // IN [1, 2]
         $notIn: [1, 2],        // NOT IN [1, 2]
         $like: '%hat',         // LIKE '%hat'
         $notLike: '%hat'       // NOT LIKE '%hat'
         $iLike: '%hat'         // ILIKE '%hat' (case insensitive) (PG only)
         $notILike: '%hat'      // NOT ILIKE '%hat'  (PG only)
         $like: { $any: ['cat', 'hat']}
         // LIKE ANY ARRAY['cat', 'hat'] - also works for iLike and notLike
         $overlap: [1, 2]       // && [1, 2] (PG array overlap operator)
         $contains: [1, 2]      // @> [1, 2] (PG array contains operator)
         $contained: [1, 2]     // <@ [1, 2] (PG array contained by operator)
         $any: [2,3]            // ANY ARRAY[2, 3]::INTEGER (PG only)

         $col: 'user.organization_id' // = "user"."organization_id", with dialect specific column identifiers, PG in this example
         */

        // Operator (=, <, >, <=, >=). Default: =

        switch(operator) {
            case '$eq':
                return '%s = ?';
            case '$gt':
                return '%s > ?';
            case '$gte':
                return '%s >= ?';
            case '$lt':
                return '%s < ?';
            case '$lte':
                return '%s <= ?';
            case '$like':
                return '%s LIKE ?';
            case '$notLike':
                return '%s NOT LIKE ?';
            case '$regex':
                return 'REGEXP_MATCH(%s, r?)';
            case '$notRegex':
                return 'NOT REGEXP_MATCH(%s, r?)';
        }

        throw new Error('operator "' + operator + '" not available in Google DataStore');
    }

    isPromise(value) {
        return value && (value instanceof Promise || (typeof value === 'object' && typeof value.then === 'function'));
    }
}

module.exports = BigQuery;