'use strict';

const Model = require('../model');
const Base = require('./base');
const Types = require('paradox-utils').SchemaTypes;
const _ = require('lodash');

const elasticsearch = require('elasticsearch');

const NotFoundError = require('../errors/not-found');
const UnknownError = require('../errors/unknown');
const PartitionError = require('../errors/partition');

class ElasticSearch extends Base {

    define(name, fields, options) {
        let model = new Model(this.conn, name, fields, options);

        fields['id'] = {
            type: Types.KEY,
            primaryKey: true
        };

        return model;
    }

    /* -------------------------------------------------------------------------
     * Connections
     * ---------------------------------------------------------------------- */

    getLink() {
        if(!this.elastic) {
            let opts = this.options.credentials;
            opts.apiVersion = '5.4';

            this.elastic = new elasticsearch.Client(opts);
        }

        return this.elastic;
    }

    /* -------------------------------------------------------------------------
     * Find all
     * ---------------------------------------------------------------------- */

    findAll(options) {
        if(options.where && options.where.id) {
            return this
                .findById(options.where.id, options)
                .then(instance => {
                    return [instance];
                });
        }

        let opts = this.extractIndexType(options);

        if(options.where && options.where.q) {
            opts.q = options.where.q;

        } else if(options.where) {
            opts.body = options.where;
        }

        opts.size = options.limit;
        opts.from = options.offset;

        if(options.order) {
            this.applyOrder(opts, options.order);
        }

        return this
            .getLink()
            .search(opts)
            .then(res => {
                return res.hits.hits.map(instance => {
                    return this.applyIdToSource(instance._id, instance._source);
                });
            });
    }

    applyOrder(opts, order) {
        if(!_.isArray(order)) {
            order = [order];
        }

        if(order.length == 2 && ['DESC', 'ASC'].indexOf(order[1]) >= 0) {
            order = [order];
        }

        let orderFields = [];

        order.forEach((subOrder) => {
            if(_.isArray(subOrder) && subOrder.length == 1) {
                orderFields.push(subOrder[0]);

            } else if(_.isArray(subOrder) && subOrder.length == 2) {
                let orderField = {};
                orderField[subOrder[0]] = subOrder[1];
                orderFields.push(orderField);

            } else if(typeof subOrder == 'string') {
                orderFields.push(subOrder);

            } else {
                throw new Error('not available order method');
            }
        });

        if(!opts.body) {
            opts.body = {};
        }

        opts.body['sort'] = orderFields;

        return this;
    }

    findById(id, options) {
        let opts = this.extractIndexType(options);
        opts.id = id;

        return this
            .getLink()
            .get(opts)
            .then(resp => {
                return this.applyIdToSource(resp._id, resp._source);
            })
            .catch(err => {
                if(err.status == 404) {
                    return Promise.reject(new NotFoundError(err));
                }

                return Promise.reject(new UnknownError(err));
            });
    }

    /* -------------------------------------------------------------------------
     * Query
     * ---------------------------------------------------------------------- */

    query(query, options) {
        let opts = this.extractIndexType(options);

        opts = Object.assign(opts, query);
        opts.size = options.limit;
        opts.from = options.offset;

        if(options.order) {
            this.applyOrder(opts, options.order);
        }

        return this
            .getLink()
            .search(opts)
            .then(res => {
                return res.hits.hits.map(instance => {
                    return this.applyIdToSource(instance._id, instance._source);
                });
            });
    }

    /* -------------------------------------------------------------------------
     * Count
     * ---------------------------------------------------------------------- */

    count(options) {
        let opts = this.extractIndexType(options);

        if(options.where && options.where.q) {
            opts.q = options.where.q;

        } else if(options.where) {
            opts.body = options.where;
        }

        return this
            .getLink()
            .count(opts)
            .then(res => {
                return res.count;
            })
            .catch(err => {
                if(err.status == 404) {
                    return 0;
                }

                return Promise.reject(new UnknownError(err));
            });
    }

    /* -------------------------------------------------------------------------
     * Modify
     * ---------------------------------------------------------------------- */

    insert(data, options) {
        return this.update(data, options);
    }

    update(data, options) {
        let opts = this.extractIndexType(options);

        if(options.where && options.where.id) {
            opts.id = options.where.id;
        }

        opts.body = data;

        return this
            .getLink()
            .index(opts)
            .then(resp => {
                return this.findById(resp._id, options);
            });
    }

    destroy(options) {
        if(options.where && options.where.id) {
            return this.destroyById(options.where.id, options);
        }

        let opts = this.extractIndexType(options);
        opts.body = options.where;
        opts.refresh = true;

        return this
            .getLink()
            .deleteByQuery(opts);
    }

    destroyById(id, options) {
        if(!id) {
            throw new Error('id is required for deleting an instance');
        }

        let opts = this.extractIndexType(options);
        opts.id = id;
        opts.refresh = true;

        return this
            .getLink()
            .delete(opts)
            .catch(err => {
                if(err.status == 404) {
                    return Promise.reject(new NotFoundError(err));
                }

                return Promise.reject(new UnknownError(err));
            })
    }

    sync(options) {
        // TODO: Crear indice con fields??
    }

    /* -------------------------------------------------------------------------
     * Utils
     * ---------------------------------------------------------------------- */

    applyIdToSource(id, source) {
        return Object.assign(source, {
            id: id,
        });
    }

    extractIndexType(options) {
        let table = options.table;

        if(!table) {
            throw new UnknownError(new Error('a table is required'));
        }

        // Namespace
        let namespace = options.namespace || this.options.namespace;

        if(!namespace) {
            throw new UnknownError(new Error('a namespace is required'));
        }

        // Partition
        if(this.options.forcePartition && (!options.partition && options.partition !== false)) {
            throw new PartitionError(true);
        }

        if(options.partition) {
            namespace += '_' + options.partition;
        }

        // Index name
        let parts = table.split(/\./);

        let indexName = namespace + '_' + parts[0];

        return {
            index: indexName,
            type: parts[1],
        };
    }
}

module.exports = ElasticSearch;