'use strict';

const chai = require('chai');
const mocha = require('mocha');
const should = require('should');
const asyncp = require('paradox-asyncp');

const Dblize = require('../index');
const _ = require('lodash');
const Promise = require('bluebird');

describe('Datastore', function() {
    this.timeout(10 * 1000);

    let conn = null;
    let TestModel = null;
    let RaceModel = null;
    let TestModelPartition = null;

    let partition = null;

    /* -------------------------------------------------------------------------
     * Before
     * ---------------------------------------------------------------------- */

    before(function() {
        conn = new Dblize({
            adapter: 'datastore',
            credentials: {
                projectId: process.env.CRED_PROJECTID,
                keyFilename: process.env.CRED_KEYFILE,
            },
            namespace: 'test',
            forcePartition: true,
            // emptyThrowException: false,
        });
    });

    /* -------------------------------------------------------------------------
     * Create
     * ---------------------------------------------------------------------- */

    const validateRecord = instance => {
        (instance).should.have.property('key').which.is.an.Number();
        (instance).should.have.property('title').which.is.an.String();
        (instance).should.have.property('episodes').which.is.an.Number();
        (instance).should.have.property('success').which.is.a.Boolean();

        if(instance.created_at) {
            (instance).should.have.property('created_at').which.is.a.Date();
            (instance).should.have.property('updated_at').which.is.a.Date();
        }

        (instance.key).should.equal(1);
        (instance.title).should.equal('this is a test');
        (instance.episodes).should.equal(4);
        (instance.success).should.be.True();
    };

    /* -------------------------------------------------------------------------
     * Indexes
     * ---------------------------------------------------------------------- */

    describe('Indexes', () => {
        it('define a model', () => {
            partition = 'demo';

            TestModel = conn
                .define('resource_data', {
                    name: {
                        type: Dblize.type.STRING,
                    },
                    array: {
                        type: Dblize.type.JSON,
                    },
                    list: {
                        type: Dblize.type.JSON,
                    },
                    nullObj: {
                        type: Dblize.type.STRING,
                    },
                    nanObj: {
                        type: Dblize.type.STRING,
                        defaultValue: NaN,
                    }
                }, {
                    indexes: [
                        {
                            name: 'test',
                            fields: ['list']
                        }
                    ]
                })
                .partition('demo');
        });

        it('should store a full indexed object', () => {
            return TestModel
                .create({
                    key: 1,
                    name: 'test',
                    array: [
                        {
                            a: 123,
                            b: 'ey',
                            c: [
                                1, 2, 3,
                            ],
                            d: NaN
                        }
                    ],
                    list: {
                        a: 123,
                        b: 'e'.repeat(1501),
                        c: [
                            1, 2, 3
                        ]
                    }
                }, {
                    fullIndex: true,
                    noIndex: ['list'],
                });
        });

        it('should store a non full indexed object', () => {
            return TestModel
                .create({
                    key: 2,
                    name: 'test',
                    array: [
                        {
                            a: 123,
                            b: 'ey',
                            c: [
                                1, 2, 3,
                            ],
                            d: NaN,
                            form: {
                                a: 222,
                                b: 'hola',
                            },
                            obj: {
                                getValues() {
                                    return {
                                        a: 2,
                                        b: 4
                                    };
                                }
                            }
                        }
                    ],
                    list: {
                        a: 123,
                        b: 'e'.repeat(1501),
                        c: [
                            1, 2, 3
                        ]
                    }
                }, {
                    fullIndex: false,
                });
        });

        it('should delete all', () => {
            return TestModel
                .destroy();
        });
    });

    /* -------------------------------------------------------------------------
     * Partitions
     * ---------------------------------------------------------------------- */

    describe('Partitions', () => {
        it('define a model', () => {
            partition = 'demo';

            TestModel = conn
                .define('resource', {
                    title: {
                        type: Dblize.type.STRING,
                    },
                    episodes: {
                        type: Dblize.type.NUMBER,
                    },
                    success: {
                        type: Dblize.type.BOOLEAN,
                    }
                });

            TestModelPartition = TestModel.partition('demo');
        });

        it('should require partition', () => {
            return TestModel
                .count()
                .should.be.rejectedWith({
                    message: 'partition not defined',
                });
        });

        it('should not require partition', () => {
            return TestModelPartition
                .count()
                .should.not.be.rejectedWith({
                    message: 'partition not defined',
                });
        });

        it('should not require partition', () => {
            return TestModel
                .count({
                    partition: false,
                })
                .should.be.fulfilledWith(0);
        });


        it('should sync the model', () => {
            return TestModel.sync({
                partition: partition
            });
        });

        it('should search one and not find nothing', () => {
            return TestModel
                .findById(1, {
                    partition: partition,
                })
                .should.be.fulfilledWith(null);
        });

        it('should search and not find nothing', () => {
            return TestModel
                .findAll({
                    partition,
                })
                .should.be.fulfilledWith([]);
        });

        it('should find and build one', () => {
            return TestModel
                .findOrBuild({
                    where: {
                        key: 1,
                    },
                    defaults: {
                        title: 'this is a test',
                        episodes: 4,
                        success: true,
                    },
                    partition: partition
                })
                .spread((instance, created) => {
                    (created).should.be.true();
                    validateRecord(instance);
                })
        });


        it('should insert a record with key 1', () => {
            return TestModel
                .create({
                    key: 1,
                    title: 'this is a test',
                    episodes: 4,
                    success: true,
                }, {
                    partition: partition
                })
                .then(instance => {
                    validateRecord(instance);
                });
        });

        it('should search one', () => {
            return TestModel
                .findById(1, {
                    partition: partition
                })
                .then(instance => {
                    validateRecord(instance);
                })
        });

        it('should search all by key and return array', () => {
            return TestModel
                .findAll({
                    where: {
                        key: 1,
                    },
                    partition: partition
                })
                .then(instances => {
                    (instances).should.be.an.Array();

                    let instance = instances[0];
                    validateRecord(instance);
                })
        });

        it('should search all by no existing key and return empty array', () => {
            return TestModel
                .findAll({
                    where: {
                        key: 123,
                    },
                    partition: partition
                })
                .should.be.fulfilledWith([]);
        });

        it('should find and not build one', () => {
            return TestModel
                .findOrBuild({
                    where: {
                        key: 1,
                    },
                    partition: partition
                })
                .spread((instance, created) => {
                    (created).should.be.false();
                    validateRecord(instance);
                })
        });

        it('should find and not create one', () => {
            return TestModel
                .findOrCreate({
                    where: {
                        key: 1,
                    },
                    partition: partition
                })
                .spread((instance, created) => {
                    (created).should.be.false();
                    validateRecord(instance);
                })
        });

        it('should search all', () => {
            return TestModel
                .findAll({
                    partition: partition,
                    order: [
                        ['title', 'DESC']
                    ]
                })
                .then(instances => {
                    let instance = instances[0];
                    validateRecord(instance);
                });
        });

        it('should query by title', () => {
            return TestModel
                .findAll({
                    where: {
                        title: 'this is a test',
                    },
                    partition: partition,
                })
                .then(instances => {
                    let instance = instances[0];
                    validateRecord(instance);
                });
        });

        it('should delete all', () => {
            return TestModel
                .destroy({
                    where: {
                        key: 1,
                    },
                    partition: partition,
                });
        });
    });

    describe('Find and update race', () => {

        before(() => {
            RaceModel = conn
                .define('race', {
                    running: {
                        type: Dblize.BOOLEAN,
                    },
                    worker: {
                        type: Dblize.STRING,
                    }
                })
                .partition(false);
        });

        it('should create the task', () => {
            return RaceModel
                .create({
                    running: false,
                });
        });

        it.skip('should race for get the task', () => {
            let workers = [];

            for(let i = 1; i <= 10; i++) {
                workers.push('worker-' + i);
            }

            return asyncp
                .eachLimit(workers, 10, workerName => {
                    return RaceModel
                        .findOneAndUpdate({
                            running: true,
                            worker: workerName,
                        }, {
                            where: {
                                running: false,
                            },
                        });
                })
                .then(results => {
                    (results.length).should.equal(1);
                });
        });

        it('should delete all', () => {
            return RaceModel
                .destroy();
        });
    });

    /* -------------------------------------------------------------------------
     * Utils
     * ---------------------------------------------------------------------- */

    describe('Utils', () => {

        let SourceResource = null;
        let RemoteResource = null;

        before(() => {
            SourceResource = TestModel.partition(false);
            RemoteResource = TestModel.partition('remote');
        });

        it('should clean all before start', () => {
            return SourceResource.destroy()
                .then(() => {
                    return RemoteResource.destroy();
                });
        });

        it('should insert data into source', () => {
            let testData = [
                {title: 'Episode 1'},
                {title: 'Episode 2'},
                {title: 'Episode 3'},
                {title: 'Episode 4'},
                {title: 'Episode 5'},
            ];

            return asyncp.eachLimit(testData, 1, data => {
                return SourceResource.create(data);
            });
        });

        it('should source contains all data', () => {
            return SourceResource
                .findAll()
                .then(records => {
                    (records).should.be.an.Array();
                    (records.length).should.equal(5);
                });
        });

        it('should remote be empty', () => {
            return RemoteResource
                .findAll()
                .then(records => {
                    (records).should.be.an.Array();
                    (records.length).should.equal(0);
                });
        });

        it('should copy all partition resources to a non partition', () => {
            return SourceResource
                .copy(RemoteResource, {
                    concurrency: 1,
                });
        });

        it('should equal records in remote and source', () => {
            return Promise.resolve(SourceResource
                .findAll())
                .delay(500) // there isn't data read consistency in cloud node
                .then(sourceRecords => {
                    return RemoteResource
                        .findAll()
                        .then(remoteRecords => {
                            (sourceRecords).should.be.an.Array();
                            (remoteRecords).should.be.an.Array();

                            (remoteRecords.length).should.equal(sourceRecords.length);

                            sourceRecords.forEach(sourceRecord => {
                                let remoteRecord = remoteRecords.find(r => r.title == sourceRecord.title);

                                let sVals = sourceRecord.getValues();
                                let rVals = remoteRecord.getValues();

                                for(let key in sVals) {
                                    if(key == 'updated_at') {
                                        continue;
                                    }

                                    let sVal = sVals[key] + "";
                                    let rVal = rVals[key] + "";

                                    (sVal).should.be.equal(rVal);
                                }
                            });
                        });
                });
        });

        after(() => {
            return SourceResource.destroy()
                .then(() => {
                    return RemoteResource.destroy();
                });
        });
    });
});